import typing as t
from html import unescape as unescape_html
import urllib.parse as up
import re

from ._exc import Unparsable
from ._struct_core import QsNode, QsRoot

t_Delimiter = t.Union[str, t.Pattern[str]]


def unpack_payload(payload: str | bytes) -> str:
    """
    Unpacks payload from bytes to string.

    Args:
        payload (str): payload to unpack

    Returns:
        str: unpacked payload
    """
    if isinstance(payload, bytes):
        return payload.decode("utf-8")
    return payload


def retrieve_charset(arg_list: list[str]) -> str | None:
    """
    Retrieve charset sentinel from list of arguments and remove it from the list

    Args:
        arg_list (list): list of arguments

    Returns:
        str: charset or None if not found
    """
    charset_arg = next((arg for arg in arg_list if arg.startswith("utf8=")), None)
    if charset_arg is None:
        return None
    # Remove charset sentinel from list
    arg_list.remove(charset_arg)
    charset_value = re.match(r"utf8=(.*)", charset_arg).group(1)
    available_charsets = ["utf-8", "iso-8859-1"]
    targeted_character = "✓"
    for charset in available_charsets:
        if unescape_html(up.unquote(charset_value, encoding=charset)) == targeted_character:
            return charset
    raise Unparsable("Unable to parse charset sentinel")


def unquote_arguments(arg_list: list[str], charset: str) -> list[str]:
    """
    Unquote arguments in list

    Args:
        arg_list (list): list of arguments
        charset (str): charset to use

    Returns:
        list: unquoted arguments
    """
    return [up.unquote(arg, encoding=charset) for arg in arg_list]


def process_numeric_entities(arg_list: list[str]) -> list[str]:
    """
    Process numeric entities in list

    Args:
        arg_list (list): list of arguments

    Returns:
        list: processed arguments
    """
    return [unescape_html(arg) for arg in arg_list]


class QsConfiguration(t.TypedDict):
    """
    Configuration options for query string parser and stringifier
    """

    charset: str
    delimiter: t_Delimiter
    allow_dots: bool
    charset_sentinel: bool
    interpret_numeric_entities: bool


class QSCore:
    """
    Base class for query string parser and stringifier
    Contains basic configuration options for parsing and stringifying query strings and
    methods for loading and preprocessing data.
    """

    config: QsConfiguration

    def __init__(
        self,
        charset: str = "utf-8",
        delimiter: t_Delimiter = "&",
        allow_dots: bool = False,
        charset_sentinel: bool = False,
        interpret_numeric_entities: bool = False,
    ):
        self.config = {
            "charset": charset,
            "delimiter": delimiter,
            "allow_dots": allow_dots,
            "charset_sentinel": charset_sentinel,
            "interpret_numeric_entities": interpret_numeric_entities,
        }

    def load_query_string(self, query_string: str | bytes) -> list[str]:
        """
        Load query string from bytes or string into list of raw arguments

        Args:
            qs (str): query string

        Returns:
            list: list of raw arguments in query string
        """
        raw_args = []
        unpacked_qs = unpack_payload(query_string)

        split_args = re.split(self.config["delimiter"], unpacked_qs)
        if self.config["charset_sentinel"]:
            charset = retrieve_charset(split_args)
            if charset is not None:
                self.config["charset"] = charset
        unquoted_args = unquote_arguments(split_args, self.config["charset"])
        if self.config["interpret_numeric_entities"]:
            raw_args = process_numeric_entities(unquoted_args)
        else:
            raw_args = unquoted_args
        return raw_args

    @staticmethod
    def load_filter_from_dict(data: dict) -> QsRoot:
        """
        Load filter from dictionary

        Args:
            data (dict): dictionary to load filter from

        Returns:
            QsRoot: root node of filter
        """
        root = QsRoot()
        for key, value in data.items():
            node = QsNode.load_from_dict(key, value)
            root.add_child(node)
        return root



from ._qs_core import QSCore
from ._struct_core import QsNode, QsRoot, t_Delimiter, NoValue
from ._exc import UnknownInputType, Unparsable, ConfigurationError

import enum
import urllib.parse as up
from copy import deepcopy


class NestedArgument:
    keys: list[str]
    value: str | NoValue

    def __init__(self):
        self.keys = []
        self.value = None

    def append_key(self, key: str):
        self.keys.append(key)

    def set_value(self, value: str):
        self.value = value

    def stringify(
        self,
        charset: str = "utf-8",
        allow_dots: bool = False,
        encode_dot_in_keys: bool = False,
        encode: bool = False,
        encode_values_only: bool = False,
    ) -> str:
        # sanitize keys to be strings
        self.keys = [str(key) for key in self.keys]
        key: str = None
        value: str | NoValue = None
        if encode_dot_in_keys:
            self.keys = [key.replace(".", "%2E") for key in self.keys]
        if allow_dots:
            key = ".".join(self.keys)
        else:
            if len(self.keys) > 1:
                key = f"{self.keys[0]}[{']['.join(self.keys[1:])}]"
            else:
                key = self.keys[0]
        if encode:
            key = f"{up.quote(key, encoding=charset, errors='xmlcharrefreplace')}"
            value = (
                f"{up.quote(self.value, encoding=charset, errors='xmlcharrefreplace')}"
                if self.value is not NoValue
                else self.value
            )
        elif encode_values_only:
            value = (
                f"{up.quote(self.value, encoding=charset, errors='xmlcharrefreplace')}"
                if self.value is not NoValue
                else self.value
            )
        else:
            value = self.value if self.value is not None else "null"
        if value is NoValue:
            return key
        else:
            return f"{key}={value}"

    def create_copy(self):
        return deepcopy(self)


class ArrayFormat(enum.Enum):
    INDICES = "indices"
    BRACKETS = "brackets"
    REPEAT = "repeat"
    COMMA = "comma"


class QsStringifier(QSCore):
    array_format: ArrayFormat
    encode: bool = True
    encode_dot_in_keys: bool = False
    encode_values_only: bool = False
    allow_empty_arrays: bool = False

    def __init__(
        self,
        charset: str = "utf-8",
        delimiter: t_Delimiter = "&",
        allow_dots: bool = False,
        encode_dot_in_keys: bool = False,
        charset_sentinel: bool = False,
        allow_empty_arrays: bool = False,
        encode: bool = True,
        encode_values_only: bool = False,
        array_format: str = "indices",
    ):
        """
        Constructor for the QsStringifier class.

        Args:
            charset (str, optional): The charset to use for encoding and decoding. Defaults to "utf-8".
            delimiter (t_Delimiter, optional): The delimiter to use for separating key-value pairs. Defaults to "&".
            allow_dots (bool, optional): Whether to allow dots in keys. Defaults to False.
            charset_sentinel (bool, optional): Whether to use the charset sentinel. Defaults to False.
            interpret_numeric_entities (bool, optional): Whether to interpret numeric entities. Defaults to False.
            array_format (str, optional): Preferred format for arrays. Defaults to "indices". If filter contains combined
                arrays and objects, the format will use preffered format where it can be applied and fallback to indices where it can't.
        """
        super().__init__(charset, delimiter, allow_dots, charset_sentinel, interpret_numeric_entities=True)
        self.encode = encode
        if encode and encode_values_only:
            raise ConfigurationError("Cannot encode values only when encoding is enabled")
        if not self.config["allow_dots"] and encode_dot_in_keys:
            raise ConfigurationError("Cannot encode dots in keys when dots are not allowed")
        self.encode_dot_in_keys = encode_dot_in_keys
        self.encode_values_only = encode_values_only
        self.allow_empty_arrays = allow_empty_arrays
        self.array_format = ArrayFormat(array_format)

    def stringify(self, filter_: dict | QsRoot) -> str:
        if isinstance(filter_, dict):
            return self.stringify_dict(filter_)
        elif isinstance(filter_, QsRoot):
            return self.stringify_tree(filter_)
        else:
            raise UnknownInputType(f"Cannot stringify {type(filter_)}, only dict or QSRoot allowed")

    def charset_sentinel_arg(self) -> NestedArgument:
        charset_sentinel = NestedArgument()
        charset_sentinel.append_key("utf8")
        charset_sentinel.set_value("✓")
        return charset_sentinel

    def stringify_tree(self, root: QsRoot) -> str:
        if root.is_empty:
            return formatted_string
        arglist: list[NestedArgument] = []
        for node in root.children:
            arglist += self.traverse_node(node, None)
        if self.config["charset_sentinel"]:
            arglist.insert(0, self.charset_sentinel_arg())
        formatted_string = self.config["delimiter"].join(
            [
                arg.stringify(
                    charset=self.config["charset"],
                    allow_dots=self.config["allow_dots"],
                    encode=self.encode,
                    encode_dot_in_keys=self.encode_dot_in_keys,
                    encode_values_only=self.encode_values_only,
                )
                for arg in arglist
            ]
        )
        return formatted_string

    def traverse_node(self, node: QsNode, parent_arg: NestedArgument) -> list[NestedArgument]:
        arg = NestedArgument() if parent_arg is None else parent_arg.create_copy()
        if node.is_leaf:
            arg.append_key(node.key)
            arg.set_value(node.value)
            return [arg]
        elif node.is_simple_array_branch:
            child_nodes = sorted(node.value, key=lambda x: x.key)
            arg.append_key(node.key)
            return self.create_array_arguments(child_nodes, arg)
        elif node.is_array_branch:
            return self.process_complex_array_branch(node, arg)
        else:
            # object branch
            arglist = []
            for child in node.value:
                if parent_arg is None:
                    arg = NestedArgument()
                else:
                    arg = parent_arg.create_copy()
                arg.append_key(node.key)
                arglist += self.traverse_node(child, arg)
            return arglist

    def create_array_arguments(self, node_list: list[QsNode], parent_arg: NestedArgument) -> list[NestedArgument]:
        """
        Create a list of NestedArgument objects from a list of QsNode (which are always array leaf nodes in correct order).

        Args:
            node_list (list[QsNode]): List of QsNode array leaf nodes in correct order (based on their key representing array index).
            parent_arg (NestedArgument): Parent argument object.

        Returns:
            list[NestedArgument]: List of NestedArgument objects.
        """
        arglist = []
        if parent_arg is None:
            raise Unparsable("Cannot create array without parent argument")
        match self.array_format:
            case ArrayFormat.INDICES:
                # iterate over the children with sorted keys
                for node in node_list:
                    arg = parent_arg.create_copy()
                    # set value of the argument to the value of the child node
                    arg.set_value(node.value)
                    # append index as key to the list of keys
                    arg.append_key(node.key)
                    arglist.append(arg)
            case ArrayFormat.BRACKETS:
                arg = parent_arg.create_copy()
                arg.set_value(f'[{",".join([node.value for node in node_list])}]')
                arglist.append(arg)
            case ArrayFormat.REPEAT:
                for node in node_list:
                    arg = parent_arg.create_copy()
                    # set value of the argument to the value of the child node
                    arg.set_value(node.value)
                    # append the current node key to the list of keys
                    arglist.append(arg)
            case ArrayFormat.COMMA:
                # same as brackets, but with commas
                arg = parent_arg.create_copy()
                arg.set_value(f'{",".join([node.value for node in node_list])}')
                arglist.append(arg)
        if arglist == []:
            if self.allow_empty_arrays:
                arg = parent_arg.create_copy()
                arg.append_key("")
                arg.set_value(NoValue)
                arglist.append(arg)
        return arglist

    def process_complex_array_branch(self, node: QsNode, parent_arg: NestedArgument) -> list[NestedArgument]:
        """
        Create a list of NestedArgument objects from a complex array branch.
        """
        # this operation is kind of tricky since we want to minimize the number of arguments
        # complex array branch means it can contain simple values but nested arrays or objects as well
        # best approach is to produce single argument that represents as many simple values as possible until first non-simple value is encountered
        # after that, 'indices' array format has to be used
        arglist = []
        simple_values = []
        for child in sorted(node.value, key=lambda x: x.key):
            if child.is_leaf:
                simple_values.append(child)
            else:
                if simple_values:
                    new_arg = parent_arg.create_copy()
                    new_arg.append_key(node.key)
                    # if there are simple values, process them first
                    arglist += self.create_array_arguments(simple_values, parent_arg)
                    simple_values = []
                # process the child node
                # handle all other items as 'indices' array format
                arg = parent_arg.create_copy()
                arg.append_key(node.key)
                arglist += self.traverse_node(child, arg)
        return arglist

    def stringify_dict(self, filter_: dict) -> str:
        root = QsRoot(None)
        for key, value in filter_.items():
            node = QsNode.load_from_dict(key, value, parse_array=True, allow_empty=True)
            root.add_child(node)
        return self.stringify_tree(root)


# shortcut method
def stringify(
    filter_: dict | QsRoot,
    charset: str = "utf-8",
    delimiter: t_Delimiter = "&",
    allow_dots: bool = False,
    encode_dot_in_keys: bool = False,
    charset_sentinel: bool = False,
    encode: bool = True,
    allow_empty_arrays: bool = False,
    encode_values_only: bool = False,
    array_format: str = "indices",
) -> str:
    """
    Shortcut method for stringifying a filter.

    Args:
        filter_ (dict | QSRoot): Filter to stringify.
        charset (str, optional): The charset to use for encoding and decoding. Defaults to "utf-8".
        delimiter (t_Delimiter, optional): The delimiter to use for separating key-value pairs. Defaults to "&".
        allow_dots (bool, optional): Whether to allow dots in keys. Defaults to False.
        charset_sentinel (bool, optional): Whether to use the charset sentinel. Defaults to False.
        interpret_numeric_entities (bool, optional): Whether to interpret numeric entities. Defaults to False.
        array_format (str, optional): Preferred format for arrays. Defaults to "indices". If filter contains combined
            arrays and objects, the format will use preffered format where it can be applied and fallback to indices where it can't.

    Returns:
        str: Stringified filter.
    """
    stringifier = QsStringifier(
        charset=charset,
        delimiter=delimiter,
        allow_dots=allow_dots,
        encode_dot_in_keys=encode_dot_in_keys,
        charset_sentinel=charset_sentinel,
        allow_empty_arrays=allow_empty_arrays,
        encode=encode,
        encode_values_only=encode_values_only,
        array_format=array_format,
    )
    return stringifier.stringify(filter_)



from .parser import parse, QsParser
from .stringifier import stringify, QsStringifier

VERSION = (1, 1, 1)
__version__ = ".".join(map(str, VERSION))

__all__ = ["parse", "QsParser", "stringify", "QsStringifier"]



import typing as t
from enum import Enum
import urllib.parse as up
from html import escape as escape_html

t_Delimiter = t.Union[str, t.Pattern[str]]


class NoValue:
    """
    Placeholder for missing values
    """

    def __str__(self) -> str:
        return "NoValue"

    def __repr__(self) -> str:
        return "NoValue"


class EnumDuplicateKeys(Enum):
    """
    Enum for duplicate keys handling
    """

    COMBINE = "combine"
    FIRST = "first"
    LAST = "last"


class QsNode:
    """
    Data structure to represent a query string as a tree for better manipulation of arrays and objects
    """

    auto_set_key: bool = False

    key: int | str | tuple
    value: t.Any | list["QsNode"]

    def __init__(self, key: str | int, value: t.Any, auto_set_key: bool = False):
        self.key = key
        self.value = value
        self.auto_set_key = auto_set_key

    @classmethod
    def load_from_dict(
        cls,
        item_key: int | str | tuple,
        item_value: t.Any,
        parse_array: bool = False,
        allow_empty: bool = False,
        auto_set_key: bool = False,
    ) -> "QsNode":
        """
        Load data (recursively) from dictionary into QSNode

        Args:
            data (dict): dictionary to load

        Returns:
            QSNode: QSNode instance
        """
        parsed_value = None
        # If item_value is a dictionary, means it's nested, or array with unknown indexes
        if isinstance(item_value, dict):
            parsed_value = []
            for key, value in item_value.items():
                new_node = cls.load_from_dict(key, value, allow_empty=allow_empty, parse_array=parse_array)
                if new_node is not None:
                    parsed_value.append(new_node)
            if not parsed_value and not allow_empty:
                raise ValueError("Empty objects are not allowed")
        # If item_value is a list, means it's an array - process as array-like-dictionary
        elif isinstance(item_value, list):
            parsed_value = [
                (
                    cls(idx, value, auto_set_key=True)
                    if not isinstance(value, (dict, list))
                    else cls.load_from_dict(
                        idx, value, parse_array=parse_array, allow_empty=allow_empty, auto_set_key=True
                    )
                )
                for idx, value in enumerate(item_value)
            ]
        else:
            if item_value is NoValue:
                return
            parsed_value = item_value
        if parse_array and item_key == "":
            return cls(0, parsed_value, auto_set_key=True)
        return cls(item_key, parsed_value, auto_set_key=auto_set_key)

    def __contains__(self, key: str | tuple) -> bool:
        """
        Check if key is present in children
        """
        if self.is_leaf:
            return False
        for child in self.value:
            if child.key == key:
                return True
        return False

    def __getitem__(self, key: str | tuple, default: t.Any = None) -> "QsNode":
        """
        Get child node by key
        """
        if self.is_leaf:
            return default
        for child in self.value:
            if child.key == key:
                return child
        return default

    def update(
        self, value: t.Any | list["QsNode"], handle_duplicate_keys: EnumDuplicateKeys = EnumDuplicateKeys.COMBINE
    ) -> None:
        """
        Update (recursively) value of current or nested node
        """
        if self.value is NoValue:
            self.value = value
            return
        if value is NoValue:
            return
        if handle_duplicate_keys == EnumDuplicateKeys.FIRST:
            return
        elif handle_duplicate_keys == EnumDuplicateKeys.LAST:
            self.value = value
            return
        else:
            # combining
            if self.is_leaf:
                self.update_leaf_case(value)
            elif self.is_array_branch:
                self.update_array_branch_case(value)
            elif self.is_object_branch:
                self.update_object_branch_case(value)
            else:
                # safety net, should never happen
                raise ValueError("Invalid node type")

    def update_leaf_case(self, value: t.Any | list["QsNode"]) -> None:
        """
        Update node if self is leaf node
        Using following behavioral expectation:
        {k: v} + {k: x} = {k: {0: v, 1: x}}
        {k: v} + {k: [x, y]} = {k: {0: v, 1: x, 2: y}}
        {k: v} + {k: {x: y}} = {k: {v: True, x: y}}
        Special case:
        {k: v} + {k: {v: w}} = {k: {v: w}}
        """
        if isinstance(value, list):
            # Scenario 2 or 3
            if all(v.is_array_node for v in value):
                # reindex incoming +1
                [v.reindex(shift=v.key + 1) for v in value]
                self.value = [QsNode(0, self.value), *value]
            else:
                # Scenario 3
                # Special case check
                if self.value in [v.key for v in value]:
                    self.value = value
                else:
                    self.value = [QsNode(self.value, True), *value]
        else:
            # Scenario 1 - based on self
            self.value = [QsNode(0, self.value), QsNode(1, value)]

    def update_array_branch_case(self, value: t.Any | list["QsNode"]) -> None:
        """
        Update node if self is array-branched node
        Using following behavioral expectation:
        {k: [v, w]} + {k: x} = {k: {0: v, 1: w, 2: x}}
        {k: [v, w]} + {k: [x, y]} = {k: {0: v, 1: w, 2: x, 3: y}}
        {k: [v, w]} + {k: {x: y}} = {k: {str(v_index): v, str(w_index): w, x: y}}
        """
        if isinstance(value, list):
            # Scenario 2
            if all(v.is_array_node for v in value):
                # reindex incoming + next index available
                for v in value:
                    if v.auto_set_key:
                        v.reindex(shift=self.value[-1].key + 1)
                    if v.key in [v.key for v in self.value]:
                        targeted_node = self[v.key]
                        targeted_node.update(v.value)
                    else:
                        self.value.append(v)

            else:
                # Scenario 3
                for child in self.value:
                    child.convert_to_object_type()
                for incoming_node in value:
                    if incoming_node.key in [v.key for v in self.value]:
                        targeted_node = self[incoming_node.key]
                        targeted_node.update(incoming_node.value)
                    else:
                        self.value.append(incoming_node)

        else:
            # Scenario 1 - append with next index available
            self.value.append(QsNode((int(self.value[-1].key) + 1), value))

    def update_object_branch_case(self, value: t.Any | list["QsNode"]) -> None:
        """
        Update node if self is object-branched node
        Using following behavioral expectation:
        {k: {v: w}} + {k: x} = {k: {v: w, x: True}}
        {k: {v: w}} + {k: [x, y]} = {k: {v: w, str(x_index): x, str(y_index): y}}
        {k: {v: w}} + {k: {x: y}} = {k: {v: w, x: y}}
        """
        if isinstance(value, list):
            if all(v.is_array_node for v in value):
                # Scenario 2
                for inc_v in value:
                    inc_v.convert_to_object_type()
                    if inc_v.key in [v.key for v in self.value]:
                        targeted_node = self[inc_v.key]
                        targeted_node.update(inc_v.value)
                    else:
                        self.value.append(inc_v)
            else:
                # Scenario 3
                for child in value:
                    if child.key in [v.key for v in self.value]:
                        targeted_node = self[child.key]
                        targeted_node.update(child.value)
                    else:
                        self.value.append(child)
        else:
            # Scenario 1
            if value in [v.key for v in self.value]:
                targeted_node = self[value]
                targeted_node.value = True
            else:
                self.value.append(QsNode(value, True))

    def reindex(self, shift: int = 0) -> None:
        """
        Reindex self if self is array node
        """
        if self.is_array_node:
            self.key += shift

    def convert_to_object_type(self) -> None:
        """
        Convert self to object type
        """
        if self.is_array_node:
            self.key = str(self.key)

    @property
    def is_leaf(self) -> bool:
        """
        Check if node is a leaf node
        """
        return not isinstance(self.value, list)

    @property
    def is_array_branch(self) -> bool:
        """
        Check if node is an array-branched node
        """
        if self.is_leaf:
            return False
        for child in self.value:
            # key is always string or tuple
            if not child.is_array_node:
                return False
        return True

    @property
    def is_simple_array_branch(self) -> bool:
        """
        Check if node is a simple array-branched node (it's values are not nested nodes and it's not a sparse array)
        """
        if self.is_array_branch and not self.is_sparse_array:
            for child in self.value:
                if isinstance(child.value, list):
                    return False
            return True
        return False

    @property
    def is_object_branch(self) -> bool:
        """
        Check if node is an object-branched node
        """
        if self.is_leaf:
            return False
        if not self.is_array_branch:
            return True
        return False

    @property
    def is_array_node(self) -> bool:
        """
        Check if node is an array leaf node
        """
        if isinstance(self.key, int):
            return True
        return False

    @property
    def is_sparse_array(self) -> bool:
        """
        Check if node is a sparse array
        """
        if self.is_array_branch:
            for idx, child in enumerate(sorted(self.value, key=lambda x: x.key)):
                if child.key != idx:
                    return True
        return False

    @property
    def is_empty_node(self) -> bool:
        """
        Check if node is empty
        """
        if self.value is NoValue or self.value == []:
            return True
        return False

    def _repr(self, level: int = 0) -> str:
        """
        Provide a nice formatted representation of the tree
        """
        level_dashes = "-" * level
        if self.is_leaf:
            return f"|{level_dashes} Leaf: ({type(self.key)}){self.key} -> {self.value}"
        else:
            formatted_string = f"|{level_dashes} Branch: {self.key}\n"
            for child in self.value:
                formatted_string += f"{child._repr(level=level + 1)}\n"
            return formatted_string

    def to_dict(self) -> dict:
        """
        Convert QsNode to dictionary
        """
        if self.is_leaf:
            return self.value
        if self.is_array_branch and not self.is_sparse_array:
            # return them in correct order
            return [child.to_dict() for child in sorted(self.value, key=lambda x: x.key)]
        return {child.key: child.to_dict() for child in self.value}

    def _preorder_traversal(self) -> t.Generator["QsNode", None, None]:
        """
        Preorder traversal of the tree
        """
        yield self
        if not self.is_leaf:
            for child in self.value:
                yield from child._preorder_traversal()

    def _postorder_traversal(self) -> t.Generator["QsNode", None, None]:
        """
        Postorder traversal of the tree
        """
        if not self.is_leaf:
            for child in self.value:
                yield from child._postorder_traversal()
        yield self

    def simplify_arrays(self):
        """
        Transform simple array-like dicts into arrays
        """
        if self.is_leaf:
            return
        if self.is_simple_array_branch:
            self.value = [child.value for child in sorted(self.value, key=lambda x: x.key)]
        else:
            for child in self.value:
                child.simplify_arrays()


class QsRoot:
    """
    Base class for query string parser and stringifier - preserves order of children
    """

    parameter_limit: int = 1000
    children: list[QsNode]

    def __init__(self, parameter_limit: int = 1000):
        self.children = []
        self.parameter_limit = parameter_limit

    def add_child(self, node: QsNode, duplicate_keys: EnumDuplicateKeys = EnumDuplicateKeys.COMBINE) -> None:
        """
        Add child node to root
        """
        if node is None:
            return
        if node.key in self:
            self[node.key].update(node.value, handle_duplicate_keys=duplicate_keys)
        else:
            if self.parameter_limit is not None and len(self.children) >= self.parameter_limit:
                return
            self.children.append(node)

    def __contains__(self, key: str) -> bool:
        """
        Check if key is present in children
        """
        for child in self.children:
            if child.key == key:
                return True
        return False

    def __getitem__(self, key: str, default: t.Any = None) -> QsNode:
        """
        Get child node by key
        """
        for child in self.children:
            if child.key == key:
                return child
        return default

    def _repr(self):
        """
        Provide a nice formatted representation of the tree
        """
        formatted_string = "Root:\n"
        for child in self.children:
            formatted_string += f"{child._repr(level=1)}\n"
        return formatted_string

    def to_dict(self) -> dict:
        """
        Convert QSRoot to dictionary
        """
        return {child.key: child.to_dict() for child in self.children}

    def _preorder_traversal(self) -> t.Generator[QsNode, None, None]:
        """
        Preorder traversal of the tree
        """
        for child in self.children:
            yield from child._preorder_traversal()

    def _postorder_traversal(self) -> t.Generator[QsNode, None, None]:
        """
        Postorder traversal of the tree
        """
        for child in self.children:
            yield from child._postorder_traversal()

    def process_array_limts(self, limit: int = None):
        """
        Process array limits
        """
        for node in self._preorder_traversal():
            if node.is_array_branch:
                # since indexing starts from 0, we need to check if max index is greater or equal to limit
                if not node.is_empty_node and limit is not None and max([child.key for child in node.value]) >= limit:
                    [child.convert_to_object_type() for child in node.value]

    def simplify_arrays(self):
        """
        Transform simple array-like dicts into arrays
        """
        for node in self.children:
            node.simplify_arrays()

    @property
    def has_sparse_arrays(self) -> bool:
        """
        Check if root has sparse arrays
        """
        for node in self._preorder_traversal():
            if node.is_sparse_array:
                return True
        return False

    @property
    def is_empty(self) -> bool:
        """
        Check if root is empty
        """
        return self.children == []



import unittest
import sys
import decimal

sys.path.append(".")


class ParserTest(unittest.TestCase):

    def test_basic_parse_objects(self):
        import src.qstion as qs

        # tests based on README of qs in npm: https://www.npmjs.com/package/qs?activeTab=readme
        # trivial query string
        self.assertDictEqual(qs.parse("foo=bar"), {"foo": "bar"})

        # single nested object
        self.assertDictEqual(qs.parse("a[b]=c"), {"a": {"b": "c"}})
        self.assertDictEqual(qs.parse("foo[bar]=baz"), {"foo": {"bar": "baz"}})

    def test_uri_encoding(self):
        import src.qstion as qs

        # test uri encoding
        self.assertDictEqual(qs.parse("a%5Bb%5D=c"), {"a": {"b": "c"}})
        self.assertDictEqual(qs.parse("a%5Bb%5D%5Bc%5D=d"), {"a": {"b": {"c": "d"}}})

    def test_object_depth(self):
        import src.qstion as qs

        # test object depth (default depth is 5)
        self.assertDictEqual(qs.parse("foo[bar][baz]=foobarbaz"), {"foo": {"bar": {"baz": "foobarbaz"}}})
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j"), {"a": {"b": {"c": {"d": {"e": {"f": {"[g][h][i]": "j"}}}}}}}
        )

        # override default depth
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=1), {"a": {"b": {"[c][d][e][f][g][h][i]": "j"}}}
        )
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=2), {"a": {"b": {"c": {"[d][e][f][g][h][i]": "j"}}}}
        )
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=3), {"a": {"b": {"c": {"d": {"[e][f][g][h][i]": "j"}}}}}
        )
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=4),
            {"a": {"b": {"c": {"d": {"e": {"[f][g][h][i]": "j"}}}}}},
        )
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=5),
            {"a": {"b": {"c": {"d": {"e": {"f": {"[g][h][i]": "j"}}}}}}},
        )
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=6),
            {"a": {"b": {"c": {"d": {"e": {"f": {"g": {"[h][i]": "j"}}}}}}}},
        )

        # test strict depth
        self.assertDictEqual(
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=1, strict_depth=False),
            {"a": {"b": {"[c][d][e][f][g][h][i]": "j"}}},
        )
        # Going over depth should raise an error
        with self.assertRaises(qs.parser.Unparsable):
            qs.parse("a[b][c][d][e][f][g][h][i]=j", max_depth=1, strict_depth=True)

    def test_parameter_limit(self):
        import src.qstion as qs

        # default limit is 1000
        self.assertDictEqual(
            qs.parse("a=1&b=2&c=3&d=4&e=5&f=6&g=7&h=8&i=9&j=10"),
            {"a": "1", "b": "2", "c": "3", "d": "4", "e": "5", "f": "6", "g": "7", "h": "8", "i": "9", "j": "10"},
        )
        # override to 2
        self.assertDictEqual(
            qs.parse("a=1&b=2&c=3&d=4&e=5&f=6&g=7&h=8&i=9&j=10", parameter_limit=2), {"a": "1", "b": "2"}
        )

    def test_custom_delimiter(self):
        import src.qstion as qs

        # delimiter can be either string or regex pattern
        self.assertDictEqual(qs.parse("a=b;c=d", delimiter=";"), {"a": "b", "c": "d"})
        self.assertDictEqual(qs.parse("a=b|c=d", delimiter=r"\|"), {"a": "b", "c": "d"})
        self.assertDictEqual(qs.parse("a=b_c=d", delimiter="_"), {"a": "b", "c": "d"})
        self.assertDictEqual(qs.parse("a=b|c=d", delimiter=r"[\|;]"), {"a": "b", "c": "d"})
        self.assertDictEqual(qs.parse("a=b;c=d", delimiter=r"[\|;]"), {"a": "b", "c": "d"})

    def test_dot_notation(self):
        import src.qstion as qs

        # allow dots
        self.assertDictEqual(qs.parse("a.b=c", allow_dots=True), {"a": {"b": "c"}})
        self.assertDictEqual(qs.parse("a.b.c=d", allow_dots=True), {"a": {"b": {"c": "d"}}})

        self.assertDictEqual(qs.parse("a.b=c", allow_dots=False), {"a.b": "c"})
        # combined syntax is forbidden
        with self.assertRaises(qs.parser.Unparsable):
            qs.parse("a[b].c=d", allow_dots=True)

        # dot notation with encoded dot in keys
        self.assertDictEqual(
            qs.parse("name%252Eobj.first=John&name%252Eobj.last=Doe", allow_dots=True, decode_dots_in_keys=True),
            {"name.obj": {"first": "John", "last": "Doe"}},
        )

        # decode dots in keys must be used with allow dots
        with self.assertRaises(qs.parser.ConfigurationError):
            qs.parse("name%252Eobj.first=John&name%252Eobj.last=Doe", decode_dots_in_keys=True)

    def test_arrays(self):
        import src.qstion as qs

        # basic array
        self.assertDictEqual(qs.parse("a[]=b", parse_arrays=True), {"a": ["b"]})
        self.assertDictEqual(qs.parse("a[]=b&a[]=c", parse_arrays=True), {"a": ["b", "c"]})

        # set indexes - must be if not set from 0, sparse arrays option must be enabled
        with self.assertRaises(qs.parser.Unparsable):
            qs.parse("a[1]=b", parse_arrays=True)
        # if using sparse arrays - result is Dictionary for clear indexing
        self.assertDictEqual(qs.parse("a[1]=b", parse_arrays=True, allow_sparse_arrays=True), {"a": {1: "b"}})
        self.assertDictEqual(qs.parse("a[1]=b&a[0]=c", parse_arrays=True), {"a": ["c", "b"]})

        # test nested arrays
        self.assertDictEqual(qs.parse("a[]=b&a[][c]=d", parse_arrays=True), {"a": ["b", {"c": "d"}]})
        # test mulidimensional arrays
        self.assertDictEqual(qs.parse("a[]=b&a[1][]=c&a[1][]=d", parse_arrays=True), {"a": ["b", ["c", "d"]]})

    def test_empty_arrays(self):
        import src.qstion as qs

        # empty arrays must be explicitly enabled
        with self.assertRaises(ValueError):
            qs.parse("a[]", parse_arrays=True)

        # empty arrays
        self.assertDictEqual(qs.parse("a[]", parse_arrays=True, allow_empty_arrays=True), {"a": []})

    def test_old_charset(self):
        import src.qstion as qs

        # old charset
        self.assertDictEqual(qs.parse("a=%A7", charset="iso-8859-1"), {"a": "§"})

    def test_charset_sentinel(self):
        import src.qstion as qs

        # charset sentinel - should be decoded as utf-8
        self.assertDictEqual(
            qs.parse("utf8=%E2%9C%93&a=%C3%B8", charset="iso-8859-1", charset_sentinel=True), {"a": "ø"}
        )

    def test_interpret_numeric_entities(self):
        import src.qstion as qs

        # interpret numeric entities
        self.assertDictEqual(qs.parse("a=%26%239786%3B"), {"a": "&#9786;"})
        self.assertDictEqual(qs.parse("a=%26%239786%3B", interpret_numeric_entities=True), {"a": "☺"})

    def test_sparse_arrays(self):
        import src.qstion as qs

        # sparse arrays - behavior differs from npm qs - sparse arrays must be explicitly enabled otherwise
        # error is raised if sparse arrays are detected,
        # also, sparse arrays are returned as dictionary with indexes as keys instead of list
        with self.assertRaises(qs.parser.Unparsable):
            qs.parse("a[1]=b", parse_arrays=True)

        self.assertDictEqual(qs.parse("a[1]=b", parse_arrays=True, allow_sparse_arrays=True), {"a": {1: "b"}})

        # if all indexes are provided, they dont have to be in correct order as long as they are consecutive
        self.assertDictEqual(
            qs.parse("a[5]=b&a[3]=c&a[1]=d&a[2]=e&a[4]=f&a[0]=g", parse_arrays=True),
            {"a": ["g", "d", "e", "c", "f", "b"]},
        )

        self.assertDictEqual(
            qs.parse("a[1]=b&a[15]=c", parse_arrays=True, allow_sparse_arrays=True), {"a": {1: "b", 15: "c"}}
        )

    def test_empty_string_as_array_value(self):
        import src.qstion as qs

        # empty string as array value
        self.assertDictEqual(qs.parse("a[]=", parse_arrays=True), {"a": [""]})

    def test_array_limit(self):
        import src.qstion as qs

        # array limit - default is 20, everything above is considered as object and whole array is converted to object
        self.assertDictEqual(qs.parse("a[]=1&a[]=2&a[]=3", parse_arrays=True), {"a": ["1", "2", "3"]})

        self.assertDictEqual(
            qs.parse("a[]=1&a[]=2&a[]=3&a[]=4", parse_arrays=True, array_limit=3),
            {"a": {"0": "1", "1": "2", "2": "3", "3": "4"}},
        )
        # over max limit by default
        self.assertDictEqual(qs.parse("a[100]=1", parse_arrays=True), {"a": {"100": "1"}})

    def test_parse_primitive(self):
        import src.qstion as qs

        # parse primitive - default is false
        # numbers are represented as decimal.Decimal
        self.assertDictEqual(qs.parse("a=1"), {"a": "1"})
        self.assertDictEqual(qs.parse("a=1", parse_primitive=True), {"a": decimal.Decimal(1)})

        # floats are represented as decimal.Decimal as well
        self.assertDictEqual(qs.parse("a=1.5"), {"a": "1.5"})
        self.assertDictEqual(qs.parse("a=1.5", parse_primitive=True), {"a": decimal.Decimal("1.5")})

        # booleas are accepted normally, with no strict case if primitive strict is false
        self.assertDictEqual(qs.parse("a=true"), {"a": "true"})
        self.assertDictEqual(qs.parse("a=true", parse_primitive=True), {"a": True})
        self.assertDictEqual(qs.parse("a=false"), {"a": "false"})
        self.assertDictEqual(qs.parse("a=false", parse_primitive=True), {"a": False})
        # use whatever case
        self.assertDictEqual(qs.parse("a=True", parse_primitive=True), {"a": True})
        self.assertDictEqual(qs.parse("a=False", parse_primitive=True), {"a": False})
        self.assertDictEqual(qs.parse("a=TRUE", parse_primitive=True), {"a": True})
        self.assertDictEqual(qs.parse("a=FALSE", parse_primitive=True), {"a": False})
        # when using primitive strict, only true/True and false/False are accepted
        self.assertDictEqual(qs.parse("a=true", parse_primitive=True, primitive_strict=True), {"a": True})
        self.assertDictEqual(qs.parse("a=false", parse_primitive=True, primitive_strict=True), {"a": False})
        self.assertDictEqual(qs.parse("a=True", parse_primitive=True, primitive_strict=True), {"a": True})
        self.assertDictEqual(qs.parse("a=False", parse_primitive=True, primitive_strict=True), {"a": False})
        # any other value is considered as string
        self.assertDictEqual(qs.parse("a=TRUE", parse_primitive=True, primitive_strict=True), {"a": "TRUE"})
        self.assertDictEqual(qs.parse("a=FALSE", parse_primitive=True, primitive_strict=True), {"a": "FALSE"})

        # null/None is accepted as well
        self.assertDictEqual(qs.parse("a=null"), {"a": "null"})
        self.assertDictEqual(qs.parse("a=null", parse_primitive=True), {"a": None})
        self.assertDictEqual(qs.parse("a=None"), {"a": "None"})
        self.assertDictEqual(qs.parse("a=None", parse_primitive=True), {"a": None})
        self.assertDictEqual(qs.parse("a=none"), {"a": "none"})
        self.assertDictEqual(qs.parse("a=none", parse_primitive=True), {"a": None})
        self.assertDictEqual(qs.parse("a=NONE", parse_primitive=True), {"a": None})
        self.assertDictEqual(qs.parse("a=NULL", parse_primitive=True), {"a": None})
        # same goes with strict primitive
        self.assertDictEqual(qs.parse("a=null", parse_primitive=True, primitive_strict=True), {"a": None})
        self.assertDictEqual(qs.parse("a=None", parse_primitive=True, primitive_strict=True), {"a": None})

    def test_comma_as_list(self):
        import src.qstion as qs

        # comma as list - default is false
        self.assertDictEqual(qs.parse("a=b,c"), {"a": "b,c"})
        self.assertDictEqual(qs.parse("a=b,c", comma=True), {"a": ["b", "c"]})

        # test combined syntax
        self.assertDictEqual(
            qs.parse("a=b,c&a[][d]=e,f&a[]=g", comma=True, parse_arrays=True), {"a": ["b", "c", {"d": ["e", "f"]}, "g"]}
        )

    def test_brackets_as_list(self):
        import src.qstion as qs

        # bracketed values as list - default is false
        self.assertDictEqual(qs.parse("a=[b,c]"), {"a": "[b,c]"})
        self.assertDictEqual(qs.parse("a=[b,c]", brackets=True), {"a": ["b", "c"]})
        # also works with quoted brackets
        self.assertDictEqual(qs.parse('a=["b","c"]', brackets=True), {"a": ["b", "c"]})
        self.assertDictEqual(qs.parse("a=[1,2]", brackets=True), {"a": ["1", "2"]})
        # also works with parse primitive
        self.assertDictEqual(qs.parse("a=[1,2]", brackets=True, parse_primitive=True), {"a": [1, 2]})
        # if parse primitive is enabled but user requires items as strings, using quotes is required
        self.assertDictEqual(qs.parse('a=["1","2"]', brackets=True, parse_primitive=True), {"a": ["1", "2"]})
        # NOTE: nesting brackets in value is not supported and will result as if separate values were provided
        # e.g.:
        self.assertDictEqual(qs.parse("a=[b,[c,d]]", brackets=True), {"a": ["b", "[c", "d]"]})

    def test_complex_query_strings(self):
        import src.qstion as qs

        # complex query strings
        self.assertDictEqual(
            qs.parse("a=b&c[0]=d&c[1]=e%3Df&f[0][0]=g&f[1][0]=h", parse_arrays=True),
            {"a": "b", "c": ["d", "e=f"], "f": [["g"], ["h"]]},
        )

        # NOTE arguments which are non-array and has no value such as query_string: `a` will not be parsed
        self.assertDictEqual(qs.parse("a&b[0]=c&b[1]=d", parse_arrays=True), {"b": ["c", "d"]})

        self.assertDictEqual(
            qs.parse("a=%82%B1%82%F1%82%C9%82%BF%82%CD%81I", charset="shift_jis"), {"a": "こんにちは！"}
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import sys

sys.path.append(".")


class TestStructuralFeatures(unittest.TestCase):

    def test_array_simplifying(self):
        import src.qstion as qs

        # set up simple filter with correct simple array inside
        data = {"foo": ["bar", "baz"]}
        parsed_root = qs.parser.QsParser.load_filter_from_dict(data)

        # verify that the filter has the correct structure
        self.assertEqual(parsed_root.children[0].key, "foo")
        # value should be list of nested Node objects
        self.assertIsInstance(parsed_root.children[0].value, list)
        # Nodes should have key indexing their position in the list and value being the string
        self.assertEqual(parsed_root.children[0].value[0].key, 0)
        self.assertEqual(parsed_root.children[0].value[0].value, "bar")
        self.assertEqual(parsed_root.children[0].value[1].key, 1)
        self.assertEqual(parsed_root.children[0].value[1].value, "baz")

        # simplifying arrays in root should transform these nested nodes into simple array
        parsed_root.simplify_arrays()
        # verify that the filter has the correct structure
        self.assertEqual(parsed_root.children[0].key, "foo")
        # value should be list of strings
        self.assertIsInstance(parsed_root.children[0].value, list)
        # strings should be the same as before
        self.assertEqual(parsed_root.children[0].value[0], "bar")
        self.assertEqual(parsed_root.children[0].value[1], "baz")

        # test with more complex structure
        data = {"foo": ["bar", {"baz": "qux"}, "quux"]}
        # this kind of array cannot be simplified since it contains a dictionary
        parsed_root = qs.parser.QsParser.load_filter_from_dict(data)
        parsed_root.simplify_arrays()
        # verify that the filter has the correct structure
        self.assertEqual(parsed_root.children[0].key, "foo")
        # value should be list of nested Node objects
        self.assertIsInstance(parsed_root.children[0].value, list)
        # Nodes should have key indexing their position in the list and value being the string
        self.assertEqual(parsed_root.children[0].value[0].key, 0)
        self.assertEqual(parsed_root.children[0].value[0].value, "bar")
        self.assertEqual(parsed_root.children[0].value[1].key, 1)
        self.assertIsInstance(parsed_root.children[0].value[1].value, list)
        self.assertEqual(parsed_root.children[0].value[1].value[0].key, "baz")
        self.assertEqual(parsed_root.children[0].value[1].value[0].value, "qux")
        self.assertEqual(parsed_root.children[0].value[2].key, 2)
        self.assertEqual(parsed_root.children[0].value[2].value, "quux")

        # test with providing different order of keys
        query_string = "foo[1]=bar&foo[2]=baz&foo[0]=qux"
        parsed_root = qs.parser.parse(query_string, parse_arrays=True, return_as_obj=True)
        parsed_root.simplify_arrays()
        # verify that the filter has the correct structure
        self.assertEqual(parsed_root.children[0].key, "foo")
        # value should be list of strings
        self.assertIsInstance(parsed_root.children[0].value, list)
        # strings should be the same as before but in correct order
        self.assertEqual(parsed_root.children[0].value[0], "qux")
        self.assertEqual(parsed_root.children[0].value[1], "bar")
        self.assertEqual(parsed_root.children[0].value[2], "baz")


if __name__ == "__main__":
    unittest.main()
